from asyncio import Lock
from datetime import datetime
from typing import TYPE_CHECKING, Any, cast

from mysqlx import Row, RowResult, Session, Table, get_session
from mysqlx.errors import OperationalError

from tg_util.src.utils import wrap_async

from .base import ArchiveBase

if TYPE_CHECKING:
    from urllib.parse import ParseResult

    from tg_util.src.tg.messages.export import MessageExport


class MySQLXArchive(ArchiveBase):
    _session: Session
    _table: Table

    def __init__(self, params: "ParseResult"):
        self._params = params
        self._lock = Lock()

    async def __aenter__(self):
        schema_name = self._params.path.strip("/")
        session = await wrap_async(
            get_session,
            user=self._params.username,
            password=self._params.password,
            host=self._params.hostname,
            port=self._params.port,
            schema=schema_name,
            use_pure=True,
        )
        self._session = await wrap_async(session.__enter__)
        schema = await wrap_async(session.get_schema, schema_name)
        self._table = await wrap_async(schema.get_table, "_archive_")
        return self

    async def __aexit__(self, *_exc: Any):
        await wrap_async(self._session.commit)
        return await wrap_async(self._session.__exit__, *_exc)

    async def prepare(self):
        return

    async def check_attr(
        self,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
    ):
        select = (
            self._table.select("msg", "hash", "downloaded")
            .where(
                "downloaded is not null and ("
                "hash = :hash or (width = :width and "
                "height = :height and size = :size and "
                "duration = :duration))"
            )
            .bind("hash", hash)
            .bind("width", width)
            .bind("height", height)
            .bind("size", size)
            .bind("duration", duration)
        )
        async with self._lock:
            result = cast("RowResult", await wrap_async(select.execute))
            if row := cast("Row | None", await wrap_async(result.fetch_one)):
                return row[0], row[1], row[2]

    async def check_id(self, file_id: int):
        async with self._lock:
            result = cast(
                "RowResult",
                await wrap_async(
                    self._table.select("msg")
                    .where("file_id = :fid and downloaded is not null")
                    .bind("fid", file_id)
                    .execute,
                ),
            )
            if row := cast("Row | None", await wrap_async(result.fetch_one)):
                return row[0]

    async def set_complete(self, file_id: int):
        async with self._lock:
            await wrap_async(
                self._table.update()
                .set("downloaded", datetime.now().isoformat(timespec="seconds"))
                .where("file_id = :fid")
                .bind("fid", file_id)
                .execute,
            )

    async def update(
        self,
        file_id: int,
        msg: str,
        msg_id: int,
        chat_id: int,
        chat_username: str | None,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
        type: str,
    ):
        insert = self._table.insert(
            "file_id",
            "msg",
            "msg_id",
            "chat_id",
            "chat_username",
            "hash",
            "width",
            "height",
            "size",
            "duration",
            "type",
        ).values(
            file_id,
            msg,
            msg_id,
            chat_id,
            chat_username,
            hash,
            width,
            height,
            size,
            duration,
            type,
        )
        async with self._lock:
            try:
                await wrap_async(insert.execute)
            except OperationalError:
                await wrap_async(
                    self._table.delete()
                    .where("file_id = :fid or hash = :hash")
                    .bind("fid", file_id)
                    .bind("hash", hash)
                    .execute,
                )
                await wrap_async(insert.execute)

    async def export(self, message: "MessageExport"):
        raise NotImplementedError



"""mysql session for telethon

changes:
- tables are not auto create
- no version table
"""

import warnings
from datetime import datetime, timezone
from typing import Any, cast

from mysqlx import Result, Row, RowResult, Schema, Session, Table, expr, get_session
from mysqlx.errors import OperationalError
from telethon.crypto import AuthKey
from telethon.sessions.memory import MemorySession, _SentFileType
from telethon.sessions.sqlite import CURRENT_VERSION as UPSTREAM_VERSION
from telethon.tl.types import (
    InputDocument,
    InputPhoto,
    PeerChannel,
    PeerChat,
    PeerUser,
    TLObject,
)
from telethon.tl.types.updates import State
from telethon.utils import get_peer_id

warnings.filterwarnings("ignore", module="mysqlx.protobuf")

CURRENT_VERSION = 7


class MySQLXSession(MemorySession):
    __session: Session
    __schema: Schema
    __tbl_entities: Table
    __tbl_sent_files: Table
    __tbl_sessions: Table
    __tbl_update_state: Table

    def __init__(
        self,
        user: str,
        password: str,
        host: str = "127.0.0.1",
        port: int = 33060,
        schema: str = "telethon",
    ):
        if CURRENT_VERSION < UPSTREAM_VERSION:
            err = (
                f"schema version {CURRENT_VERSION} is lower "
                f"than upstream's: {UPSTREAM_VERSION}"
            )
            raise RuntimeError(err)
        super().__init__()
        self.save_entities = True
        self.__session = get_session(
            user=user,
            password=password,
            host=host,
            port=port,
            schema=schema,
            use_pure=True,
        )

        self.__schema = self.__session.get_schema(schema)
        self.__tbl_entities = self.__schema.get_table("entities")
        self.__tbl_sent_files = self.__schema.get_table("sent_files")
        self.__tbl_sessions = self.__schema.get_table("sessions")
        self.__tbl_update_state = self.__schema.get_table("update_state")

        # These values will be saved
        result = cast("Row | None", self.__tbl_sessions.select().execute().fetch_one())
        if result:
            (
                self._dc_id,
                self._server_address,
                self._port,
                key,
                self._takeout_id,
            ) = result
            self._auth_key = AuthKey(data=key)

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self.__schema.name)

    def clone(self, to_instance=None):
        cloned = super().clone(to_instance)
        cloned.save_entities = self.save_entities
        return cloned

    def set_dc(self, dc_id, server_address, port):
        super().set_dc(dc_id, server_address, port)
        self._update_session_table()

        # Fetch the auth_key corresponding to this data center
        row = cast(
            "Row | None", self.__tbl_sessions.select("auth_key").execute().fetch_one()
        )
        if row and row[0]:
            self._auth_key = AuthKey(data=row[0])
        else:
            self._auth_key = None

    @MemorySession.auth_key.setter
    def auth_key(self, value):
        self._auth_key = value
        self._update_session_table()

    @MemorySession.takeout_id.setter
    def takeout_id(self, value):
        self._takeout_id = value
        self._update_session_table()

    def _update_session_table(self):
        self.__session.start_transaction()
        try:
            self.__session.sql("truncate sessions").execute()
            self.__tbl_sessions.insert(
                "dc_id",
                "server_address",
                "port",
                "auth_key",
                "takeout_id",
            ).values(
                self._dc_id,
                self._server_address,
                self._port,
                self._auth_key.key if self._auth_key else b"",
                self._takeout_id,
            ).execute()
            self.__session.commit()
        except Exception:
            self.__session.rollback()
            raise

    def get_update_state(self, entity_id: int):
        result = (
            self.__tbl_update_state.select("pts", "qts", "date", "seq")
            .where("id = :id")
            .bind("id", entity_id)
            .execute()
        )
        row = cast("Row | None", cast("RowResult", result).fetch_one())
        if row:
            pts, qts, date, seq = row
            date = datetime.fromtimestamp(date, tz=timezone.utc)
            return State(pts, qts, date, seq, unread_count=0)

    def set_update_state(self, entity_id: int, state: State):
        assert state.date
        self._insert_or_update(
            self.__tbl_update_state,
            "id",
            id=entity_id,
            pts=state.pts,
            qts=state.qts,
            date=state.date.timestamp(),
            seq=state.seq,
        )

    def get_update_states(self):  # type: ignore
        rows = cast(
            "list[Row]",
            self.__tbl_update_state.select("id", "pts", "qts", "date", "seq")
            .execute()
            .fetch_all(),
        )
        for row in rows:
            yield (
                row[0],
                State(
                    pts=row[1],
                    qts=row[2],
                    date=datetime.fromtimestamp(row[3], tz=timezone.utc),
                    seq=row[4],
                    unread_count=0,
                ),
            )

    def save(self):
        """Saves the current session object as session_user_id.session"""
        # This is a no-op if there are no changes to commit, so there's
        # no need for us to keep track of an "unsaved changes" variable.
        return self.__session.commit()

    def close(self):
        """Closes the connection unless we're working in-memory"""
        self.save()
        return self.__session.close()

    def delete(self):
        pass

    def process_entities(self, tlo: TLObject):
        """
        Processes all the found entities on the given TLObject,
        unless .save_entities is False.
        """
        if not self.save_entities:
            return

        rows = self._entities_to_rows(tlo)
        if not rows:
            return

        now = int(datetime.now().timestamp())
        for row in rows:
            self._insert_or_update(
                self.__tbl_entities,
                "id",
                id=row[0],
                hash=row[1],
                username=row[2],
                phone=row[3],
                name=row[4],
                date=now,
            )

    def get_entity_rows_by_phone(self, phone: int):  # type: ignore
        result = cast(
            "RowResult",
            self.__tbl_entities.select("id", "hash")
            .where("phone = :phone")
            .bind("phone", phone)
            .execute(),
        )
        return result.fetch_one()

    def get_entity_rows_by_username(self, username: str):
        result = cast(
            "RowResult",
            self.__tbl_entities.select("id", "hash", "date")
            .where("username = :username")
            .bind("username", username)
            .execute(),
        )
        rows = cast("list[Row]", result.fetch_all())

        if not rows:
            return None

        # If there is more than one result for the same username, evict the oldest one
        if len(rows) > 1:
            rows.sort(key=lambda t: t[2] or 0)
            update = (
                self.__tbl_entities.update().set("username", None).where("id = :id")
            )
            for row in rows[:-1]:
                update.bind("id", row[0]).execute()
        row = rows[-1]
        return row[0], row[1]

    def get_entity_rows_by_name(self, name: str):  # type: ignore
        result = cast(
            "RowResult",
            self.__tbl_entities.select("id", "hash")
            .where("name = :name")
            .bind("name", name)
            .execute(),
        )
        return result.fetch_one()

    def get_entity_rows_by_id(self, id: int, exact: bool = True):  # type: ignore
        if exact:
            select = (
                self.__tbl_entities.select("id", "hash")
                .where("id = :id")
                .bind("id", id)
            )
        else:
            select = (
                self.__tbl_entities.select("id", "hash")
                .where("id = :user or id = :chat or id = :channel")
                .bind("user", get_peer_id(PeerUser(id)))
                .bind("chat", get_peer_id(PeerChat(id)))
                .bind("channel", get_peer_id(PeerChannel(id)))
            )
        result = cast("RowResult", select.execute())
        return result.fetch_one()

    def get_file(self, md5_digest: bytes, file_size: int, cls: Any):
        result = cast(
            "RowResult",
            self.__tbl_sent_files.select("id", "hash")
            .where("md5_digest = :md5 and file_size = :size and type = :type")
            .bind("md5", md5_digest)
            .bind("size", file_size)
            .bind("type", _SentFileType.from_type(cls).value)
            .execute(),
        )
        if row := cast("Row", result.fetch_one()):
            # Both allowed classes have (id, access_hash) as parameters
            return cls(row[0], row[1])

    def cache_file(self, md5_digest: bytes, file_size: int, instance: Any):
        if not isinstance(instance, (InputDocument, InputPhoto)):
            raise TypeError("Cannot cache %s instance" % type(instance))

        self._insert_or_update(
            self.__tbl_sent_files,
            "md5_digest",
            "file_size",
            "type",
            md5_digest=md5_digest,
            file_size=file_size,
            type=_SentFileType.from_type(type(instance)).value,
            id=instance.id,
            hash=instance.access_hash,
        )

    @staticmethod
    def _insert_or_update(tbl: Table, /, *pks: str, **data: Any) -> Result:
        qpks = [f"`{pk}`" for pk in pks]
        qdata = {f"`{k}`": v for k, v in data.items()}
        try:
            return tbl.insert(*qdata.keys()).values(*qdata.values()).execute()
        except OperationalError:
            update = tbl.update()
            for k, v in qdata.items():
                if k not in qpks:
                    if v is None:
                        v = expr("null")
                    update = update.set(k, v)
            where: list[str] = []
            binds: dict[str, Any] = {}
            for n, pk in enumerate(qpks):
                where.append(f"{pk} = :val{n}")
                binds[f"val{n}"] = qdata[pk]
            return update.where(" and ".join(where)).bind(binds).execute()



from asyncio import Lock
from contextlib import asynccontextmanager
from datetime import datetime
from sqlite3 import Connection, connect

from tg_util.src.utils import wrap_async

from .base import ArchiveBase

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any
    from urllib.parse import ParseResult

    from tg_util.src.tg.messages.export import MessageExport


class SQLiteArchive(ArchiveBase):
    _conn: Connection

    def __init__(self, params: "ParseResult"):
        self._params = params
        self._lock = Lock()

    async def __aenter__(self):
        db = self._params.path
        conn = await wrap_async(
            connect,
            db,
            check_same_thread=False,
            autocommit=True,
        )
        self._conn = await wrap_async(conn.__enter__)
        if db != ":memory:":
            async with self.get_cursor() as cursor:
                await wrap_async(cursor.execute, "PRAGMA optimize")
        return self

    async def __aexit__(self, *_exc: "Any"):
        await wrap_async(self._conn.commit)
        await wrap_async(self._conn.__exit__, *_exc)

    @asynccontextmanager
    async def get_cursor(self):
        cursor = await wrap_async(self._conn.cursor)
        try:
            yield cursor
        finally:
            await wrap_async(cursor.close)

    async def prepare(self):
        async with self.get_cursor() as cursor:
            await wrap_async(
                cursor.executescript,
                "PRAGMA journal_mode = WAL; PRAGMA synchronous = NORMAL;",
            )
            await wrap_async(
                cursor.execute,
                "CREATE TABLE IF NOT EXISTS _archive_"
                "(file_id INTEGER PRIMARY KEY NOT NULL, msg TEXT NOT NULL, msg_id INTEGER NOT NULL, "
                "chat_id INTEGER NOT NULL, chat_username TEXT, hash BLOB NOT NULL, width INTEGER, "
                "height INTEGER, size INTEGER, duration REAL, downloaded TEXT DEFAULT NULL, "
                "type TEXT NOT NULL) STRICT",
            )

    async def check_attr(
        self,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
    ) -> "tuple[Any, Any, Any] | None":
        async with self._lock, self.get_cursor() as cursor:
            await wrap_async(
                cursor.execute,
                "select msg, hash, downloaded from _archive_ where "
                "downloaded is not null and (hash = ? or "
                "(width = ? and height = ? and size = ? and duration = ?))",
                (hash, width, height, size, duration),
            )
            return await wrap_async(cursor.fetchone)

    async def check_id(self, file_id: int):
        async with self._lock, self.get_cursor() as cursor:
            await wrap_async(
                cursor.execute,
                "select msg from _archive_ where file_id = ? and "
                "downloaded is not null",
                (file_id,),
            )
            if row := await wrap_async(cursor.fetchone):
                return row[0]

    async def set_complete(self, file_id: int):
        async with self._lock, self.get_cursor() as cursor:
            await wrap_async(
                cursor.execute,
                "update _archive_ set downloaded = ? where file_id = ?",
                (datetime.now().isoformat(" ", "seconds"), file_id),
            )

    async def update(
        self,
        file_id: int,
        msg: str,
        msg_id: int,
        chat_id: int,
        chat_username: str | None,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
        type: str,
    ):
        async with self._lock, self.get_cursor() as cursor:
            q = (
                "replace into _archive_ (file_id, msg, msg_id, chat_id, "
                "chat_username, hash, width, height, size, duration, type) "
                "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
            )
            await wrap_async(
                cursor.execute,
                q,
                (
                    file_id,
                    msg,
                    msg_id,
                    chat_id,
                    chat_username,
                    hash,
                    width,
                    height,
                    size,
                    duration,
                    type,
                ),
            )

    async def export(self, message: "MessageExport"):
        raise NotImplementedError



from enum import Enum

from msgspec import Struct


class Value(Struct):
    path: str
    arc: str


class FileType(Enum):
    Image = Value("Photo", "images")
    Video = Value("Video", "videos")
    Other = Value("", "files")

    @property
    def path(self):
        return self.value.path

    @property
    def arc(self):
        return self.value.arc



#!/usr/bin/env python
__version__ = "r2025.05.29-0"

import asyncio
import contextlib
import logging
import re
import sys
from argparse import ArgumentParser, BooleanOptionalAction
from enum import Enum
from pathlib import Path
from urllib.parse import ParseResult, parse_qs, urlparse

from msgspec import UNSET, Struct, UnsetType
from PIL import Image, ImageDraw
from telethon import TelegramClient
from tqdm.contrib import DummyTqdmFile
from tqdm.contrib.logging import logging_redirect_tqdm

from .src import ABC, ARGSBase, arc
from .src.input import InputFile
from .src.log import setup_logging
from .src.sheet import SheetGenerator
from .src.tg.messages.wrapper import InputMessageWrapper, MessageWrapped
from .src.tg.sessions.mysqlx import MySQLXSession
from .src.tg.utils import (
    get_file_attr,
    iter_messages,
    parse_entity,
    resolve_entity,
)
from .src.types import (
    Decodable,
    FileAlreadyExists,
    FileType,
    MessageHasNoFile,
    MessageValidationError,
)
from .src.types.tqdm import tqdm
from .src.utils import format_duration, parse_proxy, round_size, wrap_async

TYPE_CHECKING = False
if TYPE_CHECKING:
    from asyncio import _CoroutineLike
    from collections.abc import Sequence
    from typing import Any

    from telethon.hints import Entity
    from telethon.tl.custom import Message


logger = logging.getLogger(__name__)

TQDM_ERR = DummyTqdmFile(sys.stderr)
TQDM_OUT = DummyTqdmFile(sys.stdout)


class Mode(Enum):
    Interactive = 1
    File = 2
    URL = 3
    Unset = 0

    def __repr__(self):
        return self.name


class Arguments(ARGSBase):
    archive: str
    categorize: bool
    config: Path | None
    download_path: Path | None
    download_threads: int
    file: InputFile | None
    mode: Mode
    proxy: str | None
    session: str
    urls: list[tuple[str | int, int]]

    always_write_meta: bool
    create_sheet: bool
    debug: bool
    overwrite: bool
    reverse_download: bool
    single_url: bool
    thumbs_only: bool
    use_takeout: bool

    _ientity: str | int
    _imsg_id: list[tuple[int, int | None]]

    def __init__(self):
        self.mode = Mode.Unset
        self._ientity = ""
        self._imsg_id = []


class Config(Decodable):
    archive: str | UnsetType = UNSET
    categorize: bool | UnsetType = UNSET
    create_sheet: bool | UnsetType = UNSET
    debug: bool | UnsetType = UNSET
    download_path: str | UnsetType = UNSET
    download_threads: int | UnsetType = UNSET
    overwrite: bool | UnsetType = UNSET
    proxy: str | UnsetType = UNSET
    reverse_download: bool | UnsetType = UNSET
    session: str | UnsetType = UNSET
    single_url: bool | UnsetType = UNSET
    thumbs_only: bool | UnsetType = UNSET
    use_takeout: bool | UnsetType = UNSET


class DownloadResult(Struct, array_like=True):
    success: bool
    message: MessageWrapped
    context: dict[str, "Any"]


class TGDownloader(ABC):
    _archive: arc.ArchiveBase
    _args: Arguments
    _client: TelegramClient
    _client_orig: TelegramClient
    _input: InputFile
    _mode: Mode
    _sheet: SheetGenerator
    _tasks: set[asyncio.Task[DownloadResult]]
    _wait_time: float | None
    _wrapper: InputMessageWrapper

    @property
    def _loop(self):
        return self._client.loop

    def __init__(self, args: Arguments, client: TelegramClient):
        self._args = args
        self._client = client
        self._mode = args.mode
        self._wait_time = None
        self._tasks = set()
        self._archive = arc.open(urlparse(self._args.archive))
        if args.file:
            self._input = args.file
        if args.create_sheet:
            self._sheet = SheetGenerator()
        if self._args.use_takeout:
            self._wait_time = 0.0
        self._wrapper = InputMessageWrapper(
            client,
            args.download_path or Path.cwd(),
            args.categorize,
            args.create_sheet,
            args.overwrite,
            args.thumbs_only,
        )

    async def __aenter__(self):
        self._archive = await self._archive.__aenter__()
        if self._args.create_sheet:
            self._sheet = self._sheet.__enter__()
        self._client = await self._client.__aenter__()
        if self._args.use_takeout:
            self._client_orig = self._client
            self._client = await self._client.takeout().__aenter__()
        return self

    async def __aexit__(self, *_exc: "Any"):
        await self.wait_tasks()
        if self._args.use_takeout:
            await self._client.__aexit__(*_exc)
            self._client = self._client_orig
        await self._client.__aexit__(*_exc)
        await self._archive.__aexit__(*_exc)
        if self._args.create_sheet:
            self._sheet.__exit__(*_exc)

    async def run(self):
        logger.debug("current loop %s", self._loop)
        await self._archive.prepare()
        match self._mode:
            case Mode.Interactive:
                with tqdm() as prog:
                    await self.process_ids(
                        self._args._ientity,
                        self._args._imsg_id,
                        prog,
                    )
            case Mode.File:
                async with self._input.ensure_write() as f:
                    async for t in self.process_file():
                        await f.set_status(
                            t.context["lnum"],
                            "# %s" if t.success else "%s # error",
                        )
            case Mode.URL:
                with tqdm(total=len(self._args.urls), desc="Progress") as prog:
                    with tqdm() as subprog:
                        for entity, message_id in self._args.urls:
                            prog.update(1)
                            if self._args.single_url:
                                ids = (message_id, None)
                            else:
                                ids = (message_id - 1, 0)
                            await self.process_ids(entity, [ids], subprog)

    async def process_file(self):
        async for lnum, line in tqdm(aiter(self._input), "Overall", len(self._input)):
            if not (line := line.partition("#")[0].strip()):
                logger.debug("ignoring input at line %s", lnum)
                continue
            _entity, msg_id = parse_url_group(line)
            try:
                entity = await resolve_entity(self._client, _entity)
            except Exception:
                await self._input.set_status(lnum, "##%s (entity error)")
                continue

            async for message, reply_id in iter_messages(
                self._client,
                entity,
                ids=msg_id,
                wait_time=self._wait_time,
            ):
                if done := await self.add_task(
                    self.validate(message, entity, reply_id, lnum=lnum)
                ):
                    for t in done:
                        yield t
        for t in asyncio.as_completed(self._tasks):
            r = await self._handle_or_return(t)
            if r:
                yield r
        self._tasks.clear()

    async def process_ids(
        self,
        raw_entity: str | int,
        ids: list[tuple[int, int | None]],
        prog: "tqdm[Any]",
    ):
        entity = await resolve_entity(self._client, raw_entity)
        logger.debug("processing entity: %s", {raw_entity: str(entity)})
        prog.set_description(str(raw_entity))
        prog.reset(0)
        prog.update
        for start_id, end_id in ids:
            logger.debug("processing start: %s, end: %s", start_id, end_id)
            if end_id is None:
                pool = iter_messages(
                    self._client, entity, ids=start_id, wait_time=self._wait_time
                )
                prog.total += 1
            else:
                if end_id == 0:
                    last_id = getattr(
                        await anext(self._client.iter_messages(entity, limit=1)), "id"
                    )
                    logger.debug("fetching last message id got %s", last_id)
                    prog.total += last_id - start_id
                else:
                    prog.total += end_id - start_id - 1
                pool = iter_messages(
                    self._client,
                    entity,
                    min_id=start_id,
                    max_id=end_id,
                    wait_time=self._wait_time,
                    reverse=self._args.reverse_download,
                )
            async for message, reply_id in pool:
                if reply_id is None:
                    prog.update(1)
                await self.add_task(self.validate(message, entity, reply_id))
            prog.refresh()

    async def _handle_or_return(self, t: asyncio.Future[DownloadResult]):
        try:
            return await t
        except FileAlreadyExists as e:
            message, entity, file_id, meta_path = e.args
            try:
                await self._archive.set_complete(file_id)
            except Exception:
                pass
            if self._args.always_write_meta:
                await self._wrapper.write_meta(message, entity, meta_path)
        except MessageValidationError as e:
            if self._args.always_write_meta:
                await self._wrapper.write_meta(*e.args)
        except MessageHasNoFile:
            pass

    async def validate(
        self,
        message: "Message",
        entity: "Entity",
        reply_id: int | None,
        **ctx: "Any",
    ):
        entity_class, _, username, chat_id = parse_entity(entity)
        message_repr = self._wrapper.get_repr(
            message.id,
            entity_class,
            chat_id,
            username,
            reply_id,
        )
        if (file := message.file) is None:
            logger.debug("%s: message does not have any file", message_repr)
            raise MessageHasNoFile
        fattr = get_file_attr(file)
        target_path, meta_path = self._wrapper.resolve_path(
            chat_id,
            username,
            message.id,
            file.name,
            file.ext,
            reply_id,
            fattr,
        )

        if not self._wrapper.overwrite and target_path.exists():
            logger.debug(
                "%s: target file already exists, skipping download", message_repr
            )
            raise FileAlreadyExists(message, entity, fattr.id, meta_path)
        if (msg := await self._archive.check_id(fattr.id)) is not None:
            logger.debug(
                "%s: duplicate file id with message %s, skipping download",
                message_repr,
                msg,
            )
            raise MessageValidationError(message, entity, meta_path)
        wrapped = await self._wrapper.wrap(
            message,
            entity,
            fattr,
            target_path,
            meta_path,
            message_repr,
            reply_id,
        )
        match await self._archive.check_attr(
            wrapped.file_hash,
            fattr.width,
            fattr.height,
            fattr.size,
            fattr.duration,
        ):
            case msg, file_hash, downloaded:
                if not downloaded and msg == str(self):
                    pass
                elif file_hash == wrapped.file_hash:
                    logger.debug(
                        "%s: duplicate file hash with message %s, skipping download",
                        wrapped,
                        msg,
                    )
                    raise MessageValidationError(message, entity, meta_path)
                else:
                    logger.debug(
                        "%s: duplicate attribute with message %s, skipping download",
                        wrapped,
                        msg,
                    )
                    raise MessageValidationError(message, entity, meta_path)
            case None:
                await self._archive.update(
                    fattr.id,
                    str(wrapped),
                    wrapped.message.id,
                    chat_id,
                    username,
                    wrapped.file_hash,
                    fattr.width,
                    fattr.height,
                    fattr.size,
                    fattr.duration,
                    fattr.type.arc,
                )
        return await self.download_message(wrapped, **ctx)

    async def add_task(self, task: "_CoroutineLike[DownloadResult]"):
        self._tasks.add(self._loop.create_task(task))
        if len(self._tasks) >= self._args.download_threads:
            done, pending = await asyncio.wait(
                self._tasks, return_when=asyncio.FIRST_COMPLETED
            )
            self._tasks.difference_update(done)
            self._tasks.update(pending)
            return [r for t in done if (r := await self._handle_or_return(t))]
        return None

    async def wait_tasks(self):
        for t in asyncio.as_completed(self._tasks):
            await self._handle_or_return(t)
        self._tasks.clear()

    async def download_message(self, message: MessageWrapped, **ctx: "Any"):
        download_success = False
        logger.debug("downloading %s as %s", message, message.target_path.name)
        part_file = message.target_path.with_suffix(
            message.target_path.suffix + ".part"
        )
        try:
            fattr = message.file_attr
            with (
                contextlib.redirect_stderr(TQDM_ERR),  # type: ignore
                contextlib.redirect_stdout(TQDM_OUT),  # type: ignore
            ):
                if message.thumbs_only:
                    await self._client.download_media(
                        message.message,  # type: ignore
                        file=str(part_file.absolute()),
                        thumb=-1,
                    )
                else:
                    await self._client.download_media(
                        message.message,  # type: ignore
                        file=str(part_file.absolute()),
                    )
            part_file.rename(message.target_path)
            await self._wrapper.write_meta(
                message.message,
                message.entity,
                message.meta_path,
            )
            download_success = True
            await self._archive.set_complete(fattr.id)
            logger.info("%s: file downloaded", message)
            await self.post_download(message)
            return DownloadResult(download_success, message, ctx)
        except Exception:
            logger.exception("%s: download file error", message)
            return DownloadResult(download_success, message, ctx)
        finally:
            if not download_success:
                part_file.unlink(missing_ok=True)
                logger.debug("%s: uncomplete download, file deleted", message)

    async def post_download(self, message: MessageWrapped):
        fattr = message.file_attr
        if message.create_sheet:
            self._sheet.submit(message.target_path)
        if message.thumbs_only and fattr.type is FileType.Video:
            info = "%sx%s | %s | %s" % (
                fattr.width,
                fattr.height,
                round_size(fattr.size or 0),
                format_duration(fattr.duration or 0.0),
            )
            async with open_image(message.target_path) as img:
                draw = await wrap_async(ImageDraw.Draw, img)
                textbox = await wrap_async(
                    draw.textbbox,
                    (0, img.height),
                    info,
                    anchor="ls",
                )
                await wrap_async(draw.rectangle, textbox, fill=(10, 10, 10))
                await wrap_async(draw.text, (0, img.height), info, anchor="ls")
                await wrap_async(img.save, message.target_path)


@contextlib.asynccontextmanager
async def open_image(fp: Path):
    img = await wrap_async(Image.open, fp)
    try:
        yield img
    finally:
        await wrap_async(img.close)


async def main(_args: "Sequence[str] | None" = None):
    argparser, args = parse_args(_args)
    root = logging.getLogger(__package__)
    logging.root.setLevel(logging.ERROR)
    setup_logging((root,), debug=args.debug)
    logger.debug("using args: %s", args)
    match urlparse(args.session):
        case ParseResult(
            username=str(username),
            password=str(password),
            hostname=str(hostname),
            port=int(port),
            path=schema,
            query=query,
        ):
            proxy = None
            if args.proxy:
                proxy = parse_proxy(urlparse(args.proxy))
            qs = parse_qs(query)
            session = MySQLXSession(
                username, password, hostname, port, schema.strip("/")
            )
            client = TelegramClient(
                session,
                int(qs["api_id"][0], 10),
                qs["api_hash"][0],
                proxy=proxy,  # type: ignore
                catch_up=False,
                receive_updates=False,
            )
        case Never:
            argparser.error(f"invalid or incomplete session: {Never}")
    with logging_redirect_tqdm((root,)):
        async with TGDownloader(args, client) as tgdl:
            await tgdl.run()


def __main__():
    try:
        import uvloop as aio  # type: ignore
    except ImportError:
        aio = asyncio

    aio.run(main())


def parse_url_group(s: str) -> tuple[str | int, int]:
    for pat in (
        re.compile(r"((?:https?://)?t.me/\w+)/(\d+)$"),
        re.compile(r"(?:https?://)?t.me/c/(\d+)/(\d+)$"),
    ):
        if p := pat.match(s):
            entity, msg_id = p.groups()
            return int(entity) if entity.isdigit() else entity, int(msg_id)
    raise ValueError(repr(s))


def parse_args(_args: "Sequence[str] | None" = None):
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "urls",
        type=parse_url_group,
        nargs="*",
        action="store",
        metavar="URL",
    )
    misc = parser.add_argument_group("misc")
    misc.add_argument("-h", "--help", action="help", help="print this help and exit")
    misc.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable debug log",
        dest="debug",
    )
    misc.add_argument(
        "-V",
        "--version",
        action="version",
        help="print version",
        version=f"%(prog)s {__version__}",
    )
    downloads = parser.add_argument_group("downloads")
    downloads.add_argument(
        "-p",
        "--download-path",
        type=Path,
        default=None,
        help="(default: current directory)",
        metavar="PATH",
        dest="download_path",
    )
    downloads.add_argument(
        "-t",
        "--download-threads",
        type=lambda i: int(i, 10),
        default=8,
        help="(default: %(default)s)",
        metavar="NUM",
        dest="download_threads",
    )
    downloads.add_argument(
        "--categorize",
        action=BooleanOptionalAction,
        default=True,
        help="categorize downloads by chat username/id (default: %(default)s)",
        dest="categorize",
    )
    downloads.add_argument(
        "--overwrite",
        action=BooleanOptionalAction,
        default=True,
        help="overwrite downloaded files (default: %(default)s)",
        dest="overwrite",
    )
    downloads.add_argument(
        "--reverse-download",
        action=BooleanOptionalAction,
        default=False,
        help="download URL(s) in ascending order (default: %(default)s)",
        dest="reverse_download",
    )
    downloads.add_argument(
        "--single-url",
        action=BooleanOptionalAction,
        default=False,
        help="only fetch single message per URL(s) (default: %(default)s)",
        dest="single_url",
    )
    downloads.add_argument(
        "--thumbs-only",
        action=BooleanOptionalAction,
        default=False,
        help="download only thumbnails on videos (default: %(default)s)",
        dest="thumbs_only",
    )
    inputs = parser.add_argument_group("inputs")
    inputs.add_argument(
        "-i",
        "--interactive",
        action="store_const",
        help="force interactive mode",
        dest="mode",
        const=Mode.Interactive,
    )
    inputs.add_argument(
        "-f",
        type=InputFile,
        help="download URL(s) from FILE",
        dest="file",
        metavar="FILE",
    )
    options = parser.add_argument_group("options")
    options.add_argument(
        "-a",
        "--archive",
        dest="archive",
        default="sqlite::memory:",
        metavar="{sqlite,mysql}://user:pass@host:port/schema",
    )
    options.add_argument(
        "-s",
        "--session",
        metavar="mysql://user:pass@host:port/schema?api_id=id&api_hash=hash",
        dest="session",
    )
    options.add_argument(
        "-c",
        "--config",
        type=Path,
        default=None,
        help="load config from FILE",
        dest="config",
        metavar="FILE",
    )
    options.add_argument(
        "--proxy",
        default=None,
        dest="proxy",
        metavar="{http,socks4,socks5}://user:pass@host:port",
    )
    options.add_argument(
        "--takeout",
        action=BooleanOptionalAction,
        default=False,
        dest="use_takeout",
        help="use takeout session (default: %(default)s)",
    )
    post = parser.add_argument_group("post-dl")
    post.add_argument(
        "--create-sheet",
        action=BooleanOptionalAction,
        default=False,
        help="create video contact sheets on videos (default: %(default)s)",
        dest="create_sheet",
    )
    post.add_argument(
        "--meta=always",
        action="store_true",
        help="always write meta even if download fails",
        dest="always_write_meta",
    )
    args = parser.parse_args(_args, Arguments())
    if args.config:
        config = Config.decode_yaml(args.config.read_bytes())
        for sf in config.__struct_fields__:
            sv = getattr(config, sf)
            if sv is not UNSET:
                match sf:
                    case "download_path":
                        sv = Path(sv)
                setattr(args, sf, sv)
    args = parser.parse_args(_args, args)
    if args.mode is Mode.Unset:
        if args.file:
            args.mode = Mode.File
        elif args.urls:
            args.mode = Mode.URL
        else:
            args.mode = Mode.Interactive
    if args.mode is Mode.Interactive:
        entity = input("peer/entity id: ").strip()
        args._ientity = int(entity, 10) if entity.isdigit() else entity
        for s in input("message ids: ").split(","):
            match [s.strip() for s in s.partition("-")]:
                case "", "", "":
                    args._imsg_id.append((0, 0))
                case start, "-", "":
                    args._imsg_id.append((int(start) - 1, 0))
                case "", "-", end:
                    args._imsg_id.append((0, int(end) + 1))
                case start, "", "":
                    args._imsg_id.append((int(start), None))
                case start, _, end:
                    _start, _end = sorted((int(start), int(end)))
                    args._imsg_id.append((_start - 1, _end + 1))
                case never:
                    err = f"invalid input range {never!r}"
                    raise ValueError(err)
        if not args._imsg_id:
            parser.error("not enough input")
    return parser, args



from tg_util.src import ABC, abstractmethod

TYPE_CHECKING = False
if TYPE_CHECKING:
    from asyncio import Lock
    from typing import Any, Self
    from urllib.parse import ParseResult

    from tg_util.src.tg.messages.export import MessageExport


class ArchiveBase(ABC):
    _params: "ParseResult"
    _lock: "Lock"

    @abstractmethod
    def __init__(self, params: "ParseResult") -> None: ...

    @abstractmethod
    async def __aenter__(self) -> "Self": ...
    @abstractmethod
    async def __aexit__(self, *_exc: "Any") -> None: ...

    @abstractmethod
    async def prepare(self) -> None: ...

    @abstractmethod
    async def check_attr(
        self,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
    ) -> "tuple[Any, Any, Any] | None": ...
    @abstractmethod
    async def check_id(self, file_id: int) -> "Any | None": ...
    @abstractmethod
    async def set_complete(self, file_id: int) -> None: ...

    @abstractmethod
    async def update(
        self,
        file_id: int,
        msg: str,
        msg_id: int,
        chat_id: int,
        chat_username: str | None,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
        type: str,
    ) -> None: ...
    @abstractmethod
    async def export(self, message: "MessageExport") -> None: ...



__all__ = (
    "ArchiveBase",
    "open",
)

from .base import ArchiveBase

TYPE_CHECKING = False
if TYPE_CHECKING:
    from urllib.parse import ParseResult


def open(url: "ParseResult"):
    match url.scheme.lower():
        case "mysqlx":
            from .mysqlx import MySQLXArchive as arc

        case "sqlite" | "sqlite3":
            from .sqlite import SQLiteArchive as arc

        case "postgresql":
            from .psql import PSQLArchive as arc

        case Never:
            err = f"unknown database scheme: {Never}"
            raise ValueError(err)
    return arc(url)



#!/usr/bin/env python
__version__ = "r2025.05.21-3"

import logging
from argparse import ArgumentParser
from enum import Enum
from pathlib import Path
from urllib.parse import ParseResult, parse_qs, urlparse

import aiofiles
from msgspec import UNSET, UnsetType, json
from telethon import TelegramClient
from telethon.errors import ChannelPrivateError
from tqdm.contrib.logging import logging_redirect_tqdm

from .src import ABC, ARGSBase, arc
from .src.log import setup_logging
from .src.tg.messages.export import MessageExport
from .src.tg.sessions.mysqlx import MySQLXSession
from .src.tg.utils import (
    get_entity_stats,
    iter_messages,
    parse_hashtags,
    resolve_entity,
)
from .src.types import Decodable, EntityStats, tqdm
from .src.utils import parse_proxy

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any, Sequence

    from telethon.hints import Entity
    from telethon.tl.custom import Dialog

logger = logging.getLogger(__name__)


class Takeout(Enum):
    TRUE = "true"
    FALSE = "false"
    FALLBACK = "fallback"

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"

    def __str__(self):
        return self.value

    @property
    def use(self):
        return self is Takeout.TRUE or self is Takeout.FALLBACK


class Arguments(ARGSBase):
    config: Path | None
    ids: list[str | int]

    archive: str
    debug: bool
    export_path: Path
    min_ratio: float
    proxy: str | None
    session: str
    takeout: Takeout
    to_db: bool


class Config(Decodable):
    archive: str | UnsetType = UNSET
    debug: bool | UnsetType = UNSET
    export_path: str | UnsetType = UNSET
    min_ratio: float | UnsetType = UNSET
    proxy: str | UnsetType = UNSET
    session: str | UnsetType = UNSET
    takeout: Takeout | UnsetType = UNSET
    to_db: bool | UnsetType = UNSET


class TGExporter(ABC):
    _archive: arc.ArchiveBase
    _args: Arguments
    _client: TelegramClient
    _client_orig: TelegramClient
    _out: Path
    _takeout: Takeout
    _wait_time: float | None

    @property
    def _loop(self):
        return self._client.loop

    def __init__(self, args: Arguments, client: TelegramClient) -> None:
        self._args = args
        self._client = client
        self._out = args.export_path
        self._takeout = args.takeout
        self._wait_time = 0.0 if args.takeout.use else None
        if self._args.to_db:
            self._archive = arc.open(urlparse(self._args.archive))

    async def __aenter__(self):
        if self._args.to_db:
            self._archive = await self._archive.__aenter__()
        self._client = await self._client.__aenter__()
        if self._takeout.use:
            self._client_orig = self._client
            self._client = await self._client.takeout().__aenter__()
        return self

    async def __aexit__(self, *_exc: "Any"):
        if self._takeout.use:
            await self._client.__aexit__(*_exc)
            self._client = self._client_orig
        await self._client.__aexit__(*_exc)
        if self._args.to_db:
            await self._archive.__aexit__(*_exc)

    async def export_chat(
        self,
        c: TelegramClient,
        e: "Entity",
        fn: str,
        prog: tqdm["Any"],
        wait_time: float | None,
    ):
        total = 0
        async with aiofiles.open(self._out / f"{fn}.json", "wb", 0) as out:
            async for message, reply_id in iter_messages(c, e, wait_time=wait_time):
                total += 1
                if reply_id is None:
                    prog.update(1)
                message_d = message.to_dict()
                message_d["_hashtags"] = parse_hashtags(message)
                message_d["peer_id"]["_entity"] = e.to_dict()
                await out.write(json.encode(message_d) + b"\n")
                if self._args.to_db:
                    await self._archive.export(MessageExport.from_message(message))
        return total

    async def export_with_fallback(self, e: "Entity", s: EntityStats):
        fn = f"@{s.username}" if s.username else str(s.id)
        with tqdm(desc=str(s), total=s.messages) as prog:
            if (
                await self.export_chat(self._client, e, fn, prog, self._wait_time) == 0
                and self._takeout is Takeout.FALLBACK
            ):
                logger.debug(
                    "got 0 messages for %s using takeout session, "
                    "falling back to normal session",
                    s,
                )
                await self.export_chat(self._client_orig, e, fn, prog, None)

    async def export_dialogs(self, mr: float):
        td = getattr(await self._client.get_dialogs(0), "total", 0)
        dialog: Dialog
        async for dialog in tqdm(self._client.iter_dialogs(), "Dialogs", td):
            try:
                entity, stats = await resolve_entity(
                    self._client, dialog.input_entity, with_stats=True
                )
                if stats.messages == 0 and self._takeout is Takeout.FALLBACK:
                    stats = await get_entity_stats(self._client_orig, entity)
                if stats.ratio > mr:
                    logger.debug("processing %s with ratio: %.3f", stats, stats.ratio)
                    await self.export_with_fallback(entity, stats)
                else:
                    logger.debug("skipping %s with ratio: %.3f", stats, stats.ratio)
            except ChannelPrivateError:
                continue
            except Exception:
                logger.warning(
                    "skipping %s due to error", dialog.stringify(), exc_info=True
                )

    async def run(self):
        logger.debug("current loop %s", self._loop)
        match self._args.ids:
            case []:
                await self.export_dialogs(self._args.min_ratio)
            case ids:
                for i in ids:
                    try:
                        entity, stats = await resolve_entity(
                            self._client, i, with_stats=True
                        )
                        if stats.messages == 0 and self._takeout is Takeout.FALLBACK:
                            stats = await get_entity_stats(self._client_orig, entity)
                        await self.export_with_fallback(entity, stats)
                    except Exception:
                        logger.warning(
                            "skipping input: %r due to error", i, exc_info=True
                        )


async def main(_args: "Sequence[str] | None" = None):
    argparser, args = parse_args(_args)
    root = logging.getLogger(__package__)
    logging.root.setLevel(logging.ERROR)
    setup_logging((root,), debug=args.debug)
    logger.debug("using args: %s", args)
    match urlparse(args.session):
        case ParseResult(
            username=str(username),
            password=str(password),
            hostname=str(hostname),
            port=int(port),
            path=schema,
            query=query,
        ):
            proxy = None
            if args.proxy:
                proxy = parse_proxy(urlparse(args.proxy))
            qs = parse_qs(query)
            session = MySQLXSession(
                username, password, hostname, port, schema.strip("/")
            )
            client = TelegramClient(
                session,
                int(qs["api_id"][0], 10),
                qs["api_hash"][0],
                proxy=proxy,  # type: ignore
                catch_up=False,
                receive_updates=False,
            )
        case Never:
            argparser.error(f"invalid or incomplete session: {Never}")
    with logging_redirect_tqdm((root,)):
        async with TGExporter(args, client) as tgex:
            await tgex.run()


def __main__():
    try:
        import uvloop as aio  # type: ignore
    except ImportError:
        import asyncio as aio

    aio.run(main())


def parse_args(_args: "Sequence[str] | None" = None):
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "ids",
        nargs="*",
        action="store",
        type=lambda s: int(s, 10) if s.isdigit() else s,
        help="user/chat/channel id or username",
        metavar="ID",
    )
    misc = parser.add_argument_group("misc")
    misc.add_argument("-h", "--help", action="help", help="print this help and exit")
    misc.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable debug log",
        dest="debug",
    )
    misc.add_argument(
        "-V",
        "--version",
        action="version",
        help="print version",
        version=f"%(prog)s {__version__}",
    )
    exports = parser.add_argument_group("exports")
    exports.add_argument(
        "-p",
        "--export-path",
        type=Path,
        default=None,
        help="(default to current directory)",
        metavar="PATH",
        dest="export_path",
    )
    exports.add_argument(
        "--mr",
        "--min-ratio",
        type=float,
        default=0.0,
        help="minimum media to message ratio for export",
        metavar="NUM",
        dest="min_ratio",
    )
    exports.add_argument(
        "--to-db",
        action="store_true",
        help="also export to archive db",
        dest="to_db",
    )
    options = parser.add_argument_group("options")
    options.add_argument(
        "-a",
        "--archive",
        dest="archive",
        default="sqlite::memory:",
        metavar="{sqlite,mysql}://user:pass@host:port/schema",
    )
    options.add_argument(
        "-s",
        "--session",
        metavar="mysql://user:pass@host:port/schema?api_id=id&api_hash=hash",
        dest="session",
    )
    options.add_argument(
        "-c",
        "--config",
        type=Path,
        default=None,
        help="load config from FILE",
        dest="config",
        metavar="FILE",
    )
    options.add_argument(
        "--proxy",
        default=None,
        dest="proxy",
        metavar="{http,socks4,socks5}://user:pass@host:port",
    )
    options.add_argument(
        "--takeout",
        dest="takeout",
        nargs="?",
        const=Takeout.TRUE,
        default=Takeout.FALSE,
        choices=list(Takeout),
        type=Takeout,
    )
    args = parser.parse_args(_args, Arguments())
    if args.config:
        config = Config.decode_yaml(args.config.read_bytes())
        for sf in config.__struct_fields__:
            sv = getattr(config, sf)
            if sv is not UNSET:
                match sf:
                    case "export_path":
                        sv = Path(sv)
                setattr(args, sf, sv)
    args = parser.parse_args(_args, args)
    return parser, args



from asyncio import Lock
from datetime import datetime

from asyncpg import Pool, Record, UniqueViolationError, create_pool

from .base import ArchiveBase

TYPE_CHECKING = False
if TYPE_CHECKING:
    from urllib.parse import ParseResult

    from _typeshed import Unused

    from tg_util.src.tg.messages.export import MessageExport


class PSQLArchive(ArchiveBase):
    _pool: "Pool[Record]"

    def __init__(self, params: "ParseResult"):
        self._params = params
        self._pool = create_pool(self._params.geturl())
        self._lock = Lock()

    async def __aenter__(self):
        self._pool = await self._pool.__aenter__()
        return self

    async def __aexit__(self, *_exc: "Unused"):
        await self._pool.__aexit__(*_exc)

    async def prepare(self):
        q = (
            """CREATE TABLE IF NOT EXISTS "_archive_" (
            file_id bigint NOT NULL,
            msg text NOT NULL,
            msg_id int NOT NULL,
            chat_id bigint NOT NULL,
            username text NULL,
            hash char(128) NOT NULL,
            width int NULL,
            height int NULL,
            "size" bigint NULL,
            duration double precision NULL,
            downloaded timestamp NULL,
            "type" varchar(20) NOT NULL,
            CONSTRAINT "_archive__pk" PRIMARY KEY (file_id),
            CONSTRAINT hash_unique UNIQUE (hash)
            );"""
            'CREATE INDEX IF NOT EXISTS "_archive__type_IDX" ON "_archive_" ("type");'
        )
        await self._pool.execute(q)

    async def check_attr(
        self,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
    ):
        async with self._lock:
            r = await self._pool.fetchrow(
                "select msg, hash, downloaded from _archive_ where "
                "downloaded is not null and (hash = $1 or "
                "(width = $2 and height = $3 and size = $4 and duration = $5))",
                hash.hex(),
                width,
                height,
                size,
                duration,
            )
        if r:
            return r[0], r[1], r[2]

    async def check_id(self, file_id: int):
        async with self._lock:
            return await self._pool.fetchval(
                "select msg from _archive_ where file_id = $1 and downloaded is not null",
                file_id,
            )

    async def set_complete(self, file_id: int):
        async with self._lock:
            await self._pool.execute(
                "update _archive_ set downloaded = $1 where file_id = $2",
                datetime.now(),
                file_id,
            )

    async def update(
        self,
        file_id: int,
        msg: str,
        msg_id: int,
        chat_id: int,
        chat_username: str | None,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
        type: str,
    ):
        insert = (
            "insert into _archive_ (file_id, msg, msg_id, chat_id, "
            "username, hash, width, height, size, duration, type) "
            "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)",
            file_id,
            msg,
            msg_id,
            chat_id,
            chat_username,
            hash.hex(),
            width,
            height,
            size,
            duration,
            type,
        )
        async with self._lock, self._pool.acquire() as con:
            try:
                await con.execute(*insert)
            except UniqueViolationError:
                async with con.transaction():
                    await con.execute(
                        "delete from _archive_ where file_id = $1 or hash = $2",
                        file_id,
                        hash.hex(),
                    )
                    await con.execute(*insert)

    async def export(self, message: "MessageExport"):
        insert = (
            "insert into _all_chats_ values ($1, $2, $3, $4, $5, $6, $7, "
            "$8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, "
            "$20, $21, $22, $23, $24, $25, $26, $27, $28)",
            *message.as_tuple(),
        )
        async with self._pool.acquire() as con:
            try:
                await con.execute(*insert)
            except UniqueViolationError:
                async with con.transaction():
                    await con.execute(
                        "delete from _all_chats_ where chat_id = $1 and message_id = $2",
                        message.chat_id,
                        message.message_id,
                    )
                    await con.execute(*insert)



from hashlib import blake2b
from typing import TYPE_CHECKING, Any, Literal, overload

from telethon.tl.functions.upload import GetFileHashesRequest
from telethon.tl.types import (
    Channel,
    ChannelForbidden,
    Chat,
    ChatEmpty,
    ChatForbidden,
    Document,
    DocumentAttributeVideo,
    FileHash,
    InputMessagesFilterDocument,
    InputMessagesFilterPhotoVideo,
    MessageEntityHashtag,
    MessageReplies,
    PeerChannel,
    PeerChat,
    PeerUser,
    Photo,
    PhotoSizeProgressive,
    TypeInputFileLocation,
    User,
    UserEmpty,
)
from telethon.utils import get_inner_text

from tg_util.src.types import EntityStats, FileAttribute, FileType

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from telethon import TelegramClient
    from telethon.hints import Entity
    from telethon.network import MTProtoSender
    from telethon.tl.custom import Message
    from telethon.tl.custom.file import File


ENTITIES: dict[str, "Entity"] = {}
SENDERS: dict[int, "MTProtoSender"] = {}


@overload
async def resolve_entity(
    c: "TelegramClient",
    e: "Any",
    with_stats: Literal[True],
) -> tuple["Entity", EntityStats]: ...
@overload
async def resolve_entity(
    c: "TelegramClient",
    e: "Any",
    with_stats: bool = False,
) -> "Entity": ...
async def resolve_entity(c: "TelegramClient", e: "Any", with_stats: bool = False):
    if isinstance(e, str) and e.isdigit():
        e = int(e, 10)
    try:
        entity: Entity = ENTITIES[str(e)]
    except KeyError:
        if isinstance(e, int):
            try:
                entity = await c.get_entity(PeerChannel(e))  # type: ignore
            except Exception:
                try:
                    entity = await c.get_entity(PeerChat(e))  # type: ignore
                except Exception:
                    entity = await c.get_entity(PeerUser(e))  # type: ignore
        else:
            entity = await c.get_entity(e)  # type: ignore
        ENTITIES[str(e)] = entity  # type: ignore
    return (entity, await get_entity_stats(c, entity)) if with_stats else entity


async def get_file_hash(
    client: "TelegramClient",
    dc_id: int | None,
    location: TypeInputFileLocation,
):
    """raises: LocationInvalidError"""
    if dc_id and dc_id != client.session.dc_id:  # type: ignore
        try:
            sender = SENDERS[dc_id]
        except KeyError:
            sender = SENDERS[dc_id] = await client._borrow_exported_sender(dc_id)
        hashes = await client._call(sender, GetFileHashesRequest(location, 0))  # type: ignore
    else:
        hashes: list[FileHash] = await client(GetFileHashesRequest(location, 0))
    file_hash = blake2b()
    for fh in hashes:
        file_hash.update(fh.hash)
    return file_hash.digest()


def parse_hashtags(msg: "Message") -> list[str]:
    if not msg.entities:
        return []
    s = set(
        get_inner_text(
            msg.message,
            (e for e in msg.entities if isinstance(e, MessageEntityHashtag)),
        )
    )
    return sorted(s, key=str.casefold)


def parse_entity(entity: "Any") -> tuple[str, str, str, int]:
    """return: class, title, username, id"""
    entity_username = entity_name = ""
    match entity:
        case User(
            id=id,
            username=username,
            usernames=usernames,
            first_name=fname,
            last_name=lname,
        ):
            if username:
                entity_username = username
            elif usernames:
                entity_username = usernames[0].username
            entity_id = id
            entity_name = " ".join([x for x in (fname, lname) if x is not None])
        case Channel(id=id, username=username, usernames=usernames, title=title):
            if username:
                entity_username = username
            elif usernames:
                entity_username = usernames[0].username
            entity_id = id
            entity_name = title
        case UserEmpty(id=id) | ChatEmpty(id=id):
            entity_id = id
        case (
            Chat(id=id, title=title)
            | ChatForbidden(id=id, title=title)
            | ChannelForbidden(id=id, title=title)
        ):
            entity_id = id
            entity_name = title
        case _:
            raise TypeError
    return entity.__class__.__name__, entity_name, entity_username, entity_id


def get_file_attr(file: "File"):
    width = height = duration = size = None
    ftype = FileType.Other
    match file.media:
        case Document(size=sz, attributes=attributes):
            size = sz
            for attr in attributes:
                if isinstance(attr, DocumentAttributeVideo):
                    ftype = FileType.Video
                    duration = attr.duration
                    width = attr.w
                    height = attr.h
        case Photo(sizes=sizes):
            ftype = FileType.Image
            for s in sizes:
                if isinstance(s, PhotoSizeProgressive):
                    width = s.w
                    height = s.h
                    size = s.sizes[-1]
    return FileAttribute(width, height, duration, size, ftype, file.media.id)


async def get_entity_stats(c: "TelegramClient", e: "Entity"):
    type, title, username, id = parse_entity(e)
    media_count = getattr(
        await c.get_messages(e, 0, filter=InputMessagesFilterPhotoVideo),
        "total",
        0,
    )
    file_count = getattr(
        await c.get_messages(e, 0, filter=InputMessagesFilterDocument),
        "total",
        0,
    )
    message_count = getattr(await c.get_messages(e, 0), "total", -1)
    return EntityStats(
        type,
        title,
        username,
        id,
        media_count,
        file_count,
        message_count,
    )


async def iter_messages(
    client: "TelegramClient",
    entity: "Entity",
    ids: int | None = None,
    max_id: int = 0,
    min_id: int = 0,
    wait_time: float | None = None,
    reverse: bool = False,
    with_reply: bool = True,
) -> "AsyncIterator[tuple[Message, int | None]]":
    message: Message
    async for message in client.iter_messages(
        entity,
        ids=ids,  # type: ignore
        max_id=max_id,
        min_id=min_id,
        wait_time=wait_time,  # type: ignore
        reverse=reverse,
    ):
        if not message:
            continue
        yield message, None
        match with_reply, message.replies:
            case True, MessageReplies(replies=replies) if replies > 0:
                reply_id = 0
                async for reply in client.iter_messages(
                    entity,
                    wait_time=wait_time,  # type: ignore
                    reply_to=message.id,
                ):
                    if not reply:
                        continue
                    yield reply, reply_id
                    reply_id += 1



__all__ = (
    "ABC",
    "ABCMeta",
    "ARGSBase",
    "abstractmethod",
)

from abc import ABCMeta as _ABCMeta
from abc import abstractmethod

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any


class ABCMeta(_ABCMeta):
    __repr_fields__: tuple[str, ...]
    __slots__: tuple[str, ...]

    def __new__(
        cls,
        name: str,
        bases: tuple["type | ABCMeta", ...],
        namespace: dict[str, "Any"],
        /,
        **kwargs: "Any",
    ):
        if "__slots__" in namespace:
            raise TypeError("__slots__ shout not be defined")

        if "__annotations__" in namespace:
            namespace["__slots__"] = tuple(
                x for x in namespace["__annotations__"] if x not in namespace
            )
        else:
            namespace["__slots__"] = ()

        fields: list[str] = []
        for base in bases:
            try:
                fields.extend(base.__repr_fields__)
            except AttributeError:
                pass
        if fields:
            namespace["__repr_fields__"] = tuple(
                dict.fromkeys((*fields, *namespace["__slots__"]))
            )
        else:
            namespace["__repr_fields__"] = namespace["__slots__"]
        return super().__new__(cls, name, bases, namespace, **kwargs)


class ABC(metaclass=ABCMeta):
    pass


class ARGSBase(ABC):
    def __repr__(self) -> str:
        def iter_attr():
            for k in sorted(self.__repr_fields__):
                try:
                    yield k, getattr(self, k)
                except AttributeError:
                    continue

        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(["%s=%r" % attr for attr in iter_attr()]),
        )



from typing import ClassVar, Self

from msgspec import Struct, json, yaml

from .enums import FileType


class Decodable(Struct):
    __json_dec: ClassVar[json.Decoder[Self]]

    @classmethod
    def decode_json(cls, buf: bytes, /) -> Self:
        try:
            return cls.__json_dec.decode(buf)
        except AttributeError:
            cls.__json_dec = json.Decoder(cls)
            return cls.decode_json(buf)

    @classmethod
    def decode_yaml(cls, buf: bytes, /) -> Self:
        return yaml.decode(buf, type=cls)


class TLSchemaBase(Struct, tag=True, tag_field="_"):
    pass


class FileAttribute(Struct):
    width: int | None
    height: int | None
    duration: float | None
    size: int | None
    type: FileType
    id: int


class EntityStats(Struct):
    type: str
    title: str
    username: str
    id: int
    medias: int
    files: int
    messages: int

    @property
    def ratio(self):
        return (self.medias + self.files) / self.messages



import logging
from datetime import datetime

from msgspec import Struct, structs
from telethon.tl.types import (
    Document,
    DocumentAttributeFilename,
    DocumentAttributeImageSize,
    DocumentAttributeVideo,
    MessageMediaDocument,
    MessageMediaPhoto,
    Photo,
    PhotoSizeProgressive,
)
from telethon.utils import get_peer_id

from tg_util.src.tg.utils import parse_entity, parse_hashtags
from tg_util.src.utils import format_duration, round_size

TYPE_CHECKING = False
if TYPE_CHECKING:
    from telethon.tl.custom import Message

logger = logging.getLogger(__name__)


class MessageExport(Struct, array_like=True):
    date: datetime
    chat_id: int
    chat_name: str | None
    chat_username: str | None
    message_id: int
    text: str | None
    clean_text: str | None
    has_photo: bool
    has_video: bool
    size: int | None
    human_readable_size: str | None
    width: int | None
    height: int | None
    resolution: str | None
    duration: str | None
    duration_sec: float | None
    views: int | None
    forwards: int | None
    reactions: int | None
    file_name: str | None
    file_id: int | None
    file_uid: int | None
    grouped_id: int | None
    sender_id: int | None
    sender_name: str | None
    sender_username: str | None
    hashtags: str | None
    fetch_date: datetime

    @classmethod
    def from_message(cls, msg: "Message"):
        assert msg.date
        sender_name = sender_username = sender_id = None
        if msg.sender is not None:
            try:
                _, sender_name, sender_username, sender_id = parse_entity(msg.sender)
            except TypeError:
                logger.warning(
                    "unable to parse sender for entity %s: %r",
                    type(msg.sender),
                    msg.sender,
                )

        _, chat_name, chat_username, _ = parse_entity(msg.chat)
        clean_text = None
        if msg.message is not None:
            clean_text = " ".join(
                [ss for s in msg.message.splitlines() if (ss := s.strip())]
            )

        has_photo = False
        has_video = False
        size = None
        human_readable_size = None
        width = None
        height = None
        resolution = None
        duration = None
        duration_sec = None
        file_name = None
        file_id = None
        file_uid = None
        match msg.media:
            case MessageMediaDocument(document=document):
                match document:
                    case Document(
                        id=id,
                        access_hash=access_hash,
                        size=_size,
                        attributes=attributes,
                    ):
                        size = _size
                        human_readable_size = round_size(_size)
                        file_id = id
                        file_uid = access_hash
                        for a in attributes:
                            match a:
                                case DocumentAttributeVideo(
                                    w=w, h=h, duration=_duration
                                ):
                                    has_video = True
                                    duration_sec = _duration
                                    duration = format_duration(_duration)
                                    resolution = f"{w}x{h}"
                                    width = w
                                    height = h
                                case DocumentAttributeFilename(file_name=_file_name):
                                    file_name = _file_name
                                case DocumentAttributeImageSize(w=w, h=h):
                                    if not resolution:
                                        resolution = f"{w}x{h}"
                                        width = w
                                        height = h
            case MessageMediaPhoto(photo=photo):
                match photo:
                    case Photo(id=id, access_hash=access_hash, sizes=sizes):
                        has_photo = True
                        file_id = id
                        file_uid = access_hash
                        for s in sizes:
                            match s:
                                case PhotoSizeProgressive(w=w, h=h, sizes=sizes):
                                    resolution = f"{w}x{h}"
                                    width = w
                                    height = h
                                    size = sorted(sizes)[-1]
                                    human_readable_size = round_size(sorted(sizes)[-1])
        reactions = None
        if msg.reactions:
            reactions = sum(x.count for x in msg.reactions.results)

        return cls(
            msg.date.replace(tzinfo=None),
            get_peer_id(msg.chat, add_mark=False),
            chat_name,
            chat_username,
            msg.id,
            msg.message,
            clean_text,
            has_photo,
            has_video,
            size,
            human_readable_size,
            width,
            height,
            resolution,
            duration,
            duration_sec,
            msg.views,
            msg.forwards,
            reactions,
            file_name,
            file_id,
            file_uid,
            msg.grouped_id,
            sender_id,
            sender_name,
            sender_username,
            ", ".join(parse_hashtags(msg)) or None,
            datetime.now(),
        )

    def as_tuple(self):
        return structs.astuple(self)



from datetime import datetime
from typing import Any, Iterator, Self

from tg_util.src.types import Decodable, TLSchemaBase
from tg_util.src.utils import format_duration, round_size


class PeerChat(TLSchemaBase):
    chat_id: int

    @property
    def id(self):
        return self.chat_id


class PeerChannel(TLSchemaBase):
    channel_id: int

    @property
    def id(self):
        return self.channel_id


class PeerUser(TLSchemaBase):
    user_id: int

    @property
    def id(self):
        return self.user_id


class DocumentAttributeImageSize(TLSchemaBase):
    w: int
    h: int


class DocumentAttributeVideo(TLSchemaBase):
    w: int
    h: int
    duration: float


class DocumentAttributeFilename(TLSchemaBase):
    file_name: str


class Document(TLSchemaBase):
    id: int
    access_hash: int
    size: int
    attributes: list[
        DocumentAttributeImageSize
        | DocumentAttributeVideo
        | DocumentAttributeFilename
        | Any
    ]


class PhotoSizeProgressive(TLSchemaBase):
    w: int
    h: int
    sizes: list[int]


class Photo(TLSchemaBase):
    id: int
    access_hash: int
    sizes: list[PhotoSizeProgressive | Any]


class MessageMediaDocument(TLSchemaBase):
    document: Document | Any


class MessageMediaPhoto(TLSchemaBase):
    photo: Photo | Any


class ReactionCount(TLSchemaBase):
    count: int


class MessageReactions(TLSchemaBase):
    results: list[ReactionCount]

    def total_reactions(self):
        return sum(r.count for r in self.results)


class MessageElement(Decodable):
    id: int
    date: datetime
    peer_id: PeerChat | PeerChannel | PeerUser
    _hashtags: list[str]
    grouped_id: int | None = None
    media: MessageMediaDocument | MessageMediaPhoto | Any = None
    views: int | None = None
    forwards: int | None = None
    reactions: MessageReactions | None = None
    message: str = ""

    _has_photo_: bool = False
    _has_video_: bool = False
    _rounded_size_: str = ""
    _res_: str = ""
    _duration_string_: str = ""
    _duration_secs_: str = ""
    _size_: str = ""
    _file_name_: str = ""
    _file_id_: str = ""
    _file_uid_: str = ""

    def __post_init__(self) -> None:
        match self.media:
            case MessageMediaDocument(document):
                match document:
                    case Document(id, access_hash, size, attributes):
                        self._size_ = str(size)
                        self._rounded_size_ = round_size(size)
                        self._file_id_ = str(id)
                        self._file_uid_ = str(access_hash)
                        for a in attributes:
                            match a:
                                case DocumentAttributeVideo(w, h, duration):
                                    self._has_video_ = True
                                    self._duration_secs_ = str(duration)
                                    self._duration_string_ = format_duration(duration)
                                    self._res_ = f"{w}x{h}"
                                case DocumentAttributeFilename(file_name):
                                    self._file_name_ = file_name
                                case DocumentAttributeImageSize(w, h):
                                    if not self._res_:
                                        self._res_ = f"{w}x{h}"
            case MessageMediaPhoto(photo):
                match photo:
                    case Photo(id, access_hash, sizes):
                        self._has_photo_ = True
                        self._file_id_ = str(id)
                        self._file_uid_ = str(access_hash)
                        for s in sizes:
                            match s:
                                case PhotoSizeProgressive(w, h, sizes):
                                    self._res_ = f"{w}x{h}"
                                    self._size_ = str(sorted(sizes)[-1])
                                    self._rounded_size_ = round_size(sorted(sizes)[-1])

    def link(self, resolve_peer_id: bool = False):
        return f"t.me/c/{self.peer_id.id}/{self.id}"

    def clean_text(self):
        return " ".join([ss for s in self.message.splitlines() if (ss := s.strip())])

    def __iter__(self) -> Iterator[str]:
        yield self.date.replace(tzinfo=None).isoformat(" ")
        yield self.link()
        yield self.clean_text()
        yield str(self.has_photo)
        yield str(self.has_video)
        yield self._rounded_size_
        yield self._res_
        yield self._duration_string_
        yield self._duration_secs_
        yield self._size_
        yield str(self.views or 0)
        yield str(self.forwards or 0)
        if self.reactions:
            yield str(self.reactions.total_reactions())
        else:
            yield "0"
        yield self.file_name
        yield self._file_id_
        yield self._file_uid_
        yield str(self.grouped_id or "")

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, v: Self):
        return self._natsort < v._natsort

    def __le__(self, v: Self):
        return self._natsort <= v._natsort

    def __gt__(self, v: Self):
        return self._natsort > v._natsort

    def __ge__(self, v: Self):
        return self._natsort >= v._natsort

    @property
    def _natsort(self):
        return (self.peer_id.id, self.id)

    @property
    def file_name(self):
        return self._file_name_

    @property
    def has_photo(self):
        return self._has_photo_

    @property
    def has_video(self):
        return self._has_video_



import os
import sys
from pathlib import Path

from telethon.sessions import SQLiteSession


class CustomSession(SQLiteSession):
    def __init__(self, name: str) -> None:
        self.__orig_session = Path(sys.argv[0]).with_name(name).with_suffix(".session")
        if not self.__orig_session.exists():
            SQLiteSession(str(self.__orig_session)).close()

        if sys.platform == "win32":
            root = Path("R:\\")
        else:
            root = Path("/tmp")
        self.__temp_session = root / f"{name}_{os.getpid()}.session"
        with (
            open(self.__temp_session, "wb", 0) as fo,
            open(self.__orig_session, "rb", 0) as fi,
        ):
            fo.writelines(fi)
        super().__init__(str(self.__temp_session))

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self.filename)

    def close(self):
        super().close()
        self.__orig_session.write_bytes(self.__temp_session.read_bytes())
        with (
            open(self.__orig_session, "wb", 0) as fo,
            open(self.__temp_session, "rb", 0) as fi,
        ):
            fo.writelines(fi)
        self.delete()



import logging

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterable

logFormat = logging.Formatter(
    r"%(asctime)s: %(module)s.%(funcName)s: %(levelname)s: %(message)s"
)
logHandler = logging.StreamHandler()
logHandler.setFormatter(logFormat)


def setup_logging(
    loggers: "Iterable[logging.Logger | str]",
    handler: logging.Handler | None = None,
    debug: bool = False,
):
    if not handler:
        handler = logHandler
    for logger in loggers:
        if isinstance(logger, str):
            logger = logging.getLogger(logger)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG if debug else logging.INFO)



import logging
from hashlib import blake2b
from typing import TYPE_CHECKING, Any, cast

from msgspec import Struct, json
from telethon.errors import FileReferenceExpiredError
from telethon.utils import get_input_location

from tg_util.src.tg.utils import get_file_hash, parse_hashtags
from tg_util.src.types import FileAttribute, FileType, MessageHasNoFile
from tg_util.src.utils import wrap_async

if TYPE_CHECKING:
    from pathlib import Path

    from telethon import TelegramClient
    from telethon.hints import Entity
    from telethon.tl.custom import Message

logger = logging.getLogger(__name__)


class InputMessageWrapper(Struct):
    client: "TelegramClient"
    dl_path: "Path"
    categorize: bool
    create_sheet: bool
    overwrite: bool
    thumbs_only: bool

    async def wrap(
        self,
        message: "Message",
        entity: "Entity",
        file_attr: FileAttribute,
        target_path: "Path",
        meta_path: "Path",
        message_repr: str,
        reply_id: int | None,
    ):
        message, hash = await self.get_file_hash(message, message_repr)
        return MessageWrapped(
            message,
            entity,
            reply_id,
            file_attr,
            hash,
            target_path,
            meta_path,
            (
                self.create_sheet
                if file_attr.type is FileType.Video and not self.thumbs_only
                else False
            ),
            self.overwrite,
            self.thumbs_only,
            message_repr,
        )

    def resolve_path(
        self,
        chat_id: int,
        chat_username: str,
        message_id: int,
        file_name: Any,
        file_ext: Any,
        reply_id: int | None,
        file_attr: FileAttribute,
    ):
        """return target_path, meta_path"""
        base_name = "%s_%s%s%s" % (
            chat_id,
            message_id,
            f"_{reply_id}" if reply_id is not None else "",
            file_ext or "",
        )
        if file_attr.type is FileType.Other:
            base_name = file_name or base_name
        if not self.categorize:
            dl_path = self.dl_path
        elif chat_username:
            dl_path = self.dl_path / f"@{chat_username}"
        else:
            dl_path = self.dl_path / str(chat_id)
        if self.thumbs_only:
            dl_path = dl_path.with_name(dl_path.name + " - thumbs")
        meta_path = (dl_path / "Meta" / base_name).with_suffix(".json")
        target_path = dl_path / file_attr.type.path / base_name
        if self.thumbs_only and file_attr.type is FileType.Video:
            target_path = target_path.with_suffix(".webp")
        return target_path, meta_path

    async def get_file_hash(self, message: "Message", message_repr: str):
        try:
            dc, loc = get_input_location(message.media)
            hash = await get_file_hash(self.client, dc, loc)
        except FileReferenceExpiredError:
            message = await self.refetch(message)
            if not (file := message.file):
                raise MessageHasNoFile
            try:
                dc, loc = get_input_location(file.media)
                hash = await get_file_hash(self.client, dc, loc)
            except Exception:
                logger.warning(
                    "unable to get file hash for %s", message_repr, exc_info=True
                )
                hash = blake2b(file.media.id.to_bytes(8)).digest()
        except TypeError:
            logger.warning(
                "unable to get file hash for %s", message_repr, exc_info=True
            )
            assert message.file
            hash = blake2b(message.file.media.id.to_bytes(8)).digest()
        return message, hash

    def get_repr(
        self,
        message_id: int,
        entity_class: str,
        entity_id: int | None,
        username: str | None,
        reply_id: int | None,
    ):
        return "Message(id=%r, reply_id=%r, from=%s" % (
            message_id,
            reply_id,
            "%s(%s)"
            % (
                entity_class,
                ", ".join(
                    [
                        f"{x}={y!r}"
                        for x, y in (
                            ("id", entity_id),
                            ("username", username),
                        )
                        if y is not None
                    ]
                ),
            ),
        )

    async def refetch(self, message: "Message"):
        return cast(
            "Message",
            await self.client.get_messages(message.input_chat, ids=message.id),
        )

    async def write_meta(self, message: "Message", entity: "Entity", fp: "Path"):
        meta = message.to_dict()
        meta["_hashtags"] = parse_hashtags(message)
        meta["peer_id"]["_entity"] = entity.to_dict()
        await wrap_async(fp.parent.mkdir, parents=True, exist_ok=True)
        await wrap_async(
            fp.write_bytes,
            await wrap_async(
                json.format, await wrap_async(json.encode, meta, enc_hook=str)
            ),
        )


class MessageWrapped(Struct):
    message: "Message"
    entity: "Entity"
    reply_id: int | None

    file_attr: FileAttribute
    file_hash: bytes
    target_path: "Path"
    meta_path: "Path"
    create_sheet: bool
    overwrite: bool
    thumbs_only: bool

    _str_: str

    def __str__(self):
        return self._str_



from asyncio.events import get_running_loop
from datetime import timedelta
from functools import partial
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, cast, overload

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from collections.abc import Awaitable, Callable
    from urllib.parse import ParseResult


def round_size(n: float | int) -> str:
    for unit in ("Bytes", "KB", "MB"):
        if n < 1024.0:
            if unit == "Bytes":
                return f"{n} {unit}"
            else:
                return f"{n:.1f} {unit}"
        n /= 1024.0
    return f"{n:.1f} GB"


def format_duration(secs: float) -> str:
    return str(timedelta(seconds=secs))


def parse_proxy(url: "ParseResult"):
    from python_socks import ProxyType

    proxy: dict[str, Any] = {}
    match url.scheme.lower():
        case "socks" | "socks5":
            proxy["proxy_type"] = ProxyType.SOCKS5
        case "socks4":
            proxy["proxy_type"] = ProxyType.SOCKS4
        case "http" | "https":
            proxy["proxy_type"] = ProxyType.HTTP
    proxy["addr"] = url.hostname
    proxy["port"] = url.port
    if url.username:
        proxy["username"] = url.username
    if url.password:
        proxy["password"] = url.password
    return proxy


_loop: "AbstractEventLoop"


@overload
async def wrap_async[**P, T](
    func: "Callable[P, T]",
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T: ...
@overload
async def wrap_async[**P, T](
    func: "Callable[P, Awaitable[T]]",
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T: ...
async def wrap_async[**P, T](
    func: "Callable[P, T | Awaitable[T]]",
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if iscoroutinefunction(func):
        return await func(*args, **kwargs)
    if callable(func):
        func = cast("Callable[P, T]", func)
        global _loop
        try:
            loop = _loop
        except NameError:
            loop = _loop = get_running_loop()
        return await loop.run_in_executor(None, partial(func, *args, **kwargs))
    err = f"{func} is neither a callable or awaitable"
    raise TypeError(err)



from asyncio import Lock
from contextlib import asynccontextmanager
from pathlib import Path

from .types import ABC
from .utils import wrap_async


class InputFile(ABC):
    _fpath: Path
    _write_lock: Lock
    content: dict[int, tuple[Lock, str]]

    def __init__(self, fp: str):
        self._fpath = Path(fp)
        self.content = {}
        with open(self._fpath, "r", encoding="utf-8") as f:
            self.content = {n: (Lock(), line.strip()) for n, line in enumerate(f, 1)}
        self._write_lock = Lock()

    def __repr__(self):
        return f"{self.__class__.__name__}({self._fpath})"

    def __len__(self) -> int:
        return self.content.__len__()

    async def __aiter__(self):
        for x in sorted(self.content):
            yield x, await self.get(x)

    @asynccontextmanager
    async def ensure_write(self):
        try:
            yield self
        finally:
            await self.write()

    async def set(self, k: int, v: str):
        try:
            lock = self.content[k][0]
        except KeyError:
            lock = Lock()
        async with lock:
            self.content[k] = (lock, v)

    async def get(self, k: int):
        async with self.content[k][0]:
            return self.content[k][1]

    async def set_status(self, at: int, fmt: str):
        await self.set(at, fmt % await self.get(at))

    async def write(self):
        async with self._write_lock:
            await wrap_async(
                self._fpath.write_text,
                "".join(["%s\n" % await self.get(n) for n in sorted(self.content)]),
                encoding="utf-8",
            )



__all__ = (
    "ABC",
    "ABCMeta",
    "ARGSBase",
    "Decodable",
    "EntityStats",
    "FileAlreadyExists",
    "FileAttribute",
    "FileType",
    "MessageHasNoFile",
    "MessageValidationError",
    "TLSchemaBase",
    "abstractmethod",
    "tqdm",
)

from ._abc import ABC, ABCMeta, ARGSBase, abstractmethod
from .enums import FileType
from .errors import FileAlreadyExists, MessageHasNoFile, MessageValidationError
from .structs import Decodable, EntityStats, FileAttribute, TLSchemaBase
from .tqdm import tqdm



import logging
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait as wait_futures
from pathlib import Path
from threading import Event, Lock, Thread

from .types import ABC

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Concatenate, Self

    from _typeshed import Unused

logger = logging.getLogger(__name__)


class SheetGenerator(ABC):
    _closed: Event
    _finalizing: Event
    _running: Event
    _stopped: Event
    lock: Lock
    pool: ThreadPoolExecutor
    queue: list[Path]
    thread: Thread

    @staticmethod
    def __ensure_open[T, **PT](func: "Callable[Concatenate[SheetGenerator, PT], T]"):
        def wrapper(self: "Self", *args: PT.args, **kwargs: PT.kwargs) -> T:
            if self._closed.is_set():
                raise RuntimeError
            return func(self, *args, **kwargs)

        return wrapper

    def __init__(self):
        self._closed = Event()
        self._finalizing = Event()
        self._running = Event()
        self._stopped = Event()
        self.lock = Lock()
        self.queue = []

    def __enter__(self):
        return self

    def __exit__(self, *_: "Unused"):
        return self.close()

    @__ensure_open
    def submit(self, val: Path):
        logger.debug("submitting %s for sheet generation", val.name)
        with self.lock:
            self.queue.append(val)
        if not self._running.is_set():
            logger.debug("thread is not started, starting thread")
            self.thread = Thread(target=self._run, daemon=True)
            self.thread.start()

    @__ensure_open
    def submits(self, vals: "Iterable[Path]"):
        with self.lock:
            self.queue.extend(vals)
        if not self._running.is_set():
            self.thread = Thread(target=self._run, daemon=True)
            self.thread.start()

    def get(self):
        with self.lock:
            return self.queue.pop(0)

    @__ensure_open
    def close(self):
        self._closed.set()
        logger.debug("queue closed")
        if self._running.is_set():
            logger.debug("waiting thread")
            self.thread.join()
        if not self._stopped.is_set() and self.queue:
            self._finalize()
        self.queue.clear()

    def stop(self):
        self._stopped.set()
        if self._finalizing.is_set():
            return self.pool.shutdown()

    def _finalize(self):
        try:
            logger.debug("finalizing sheet generation")
            self._finalizing.set()
            with ThreadPoolExecutor(4) as self.pool:
                wait_futures(self.pool.submit(generate_sheet, p) for p in self.queue)
        finally:
            self._finalizing.clear()

    def _run(self):
        try:
            self._running.set()
            logger.debug("running flag set")
            logger.debug(
                "is closed: %s, is stopped: %s",
                self._closed.is_set(),
                self._stopped.is_set(),
            )
            while not self._closed.is_set() and not self._stopped.is_set():
                try:
                    generate_sheet(self.get())
                except IndexError:
                    break
        finally:
            self._running.clear()
            logger.debug("running flag cleared")


def generate_sheet(file: Path):
    out_path = file.parent / "sheets"
    out_path.mkdir(parents=True, exist_ok=True)
    logger.debug("%s: generating sheet", file.name)
    p = subprocess.run(
        (
            "vcsi.py",
            "-t",
            *("-o", out_path / (file.name + ".webp")),
            *("-w", "4000"),
            *("-g", "6x6"),
            *("-f", "webp"),
            *("--timestamp-font-size", "20"),
            *("--metadata-font-size", "30"),
            *("--delay-percent", "0"),
            "--",
            file,
        ),
        shell=sys.platform == "win32",
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    if p.returncode == 0:
        logger.info("%s: generating sheet success", file.name)
    else:
        logger.warning("%s: generating sheet failed", file.name)
